--[[
Drathals HUD2 (c) 2007 by Horacio Hoyos / Arcanefoam
Drathals HUD (c) 2006 by Caeryn Dryad
Drathals HUD (c) 2006 by Markus Inger / Drathal / Silberklinge / Silbersegen

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are 
met:

    * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above 
       copyright notice, this list of conditions and the following 
       disclaimer in the documentation and/or other materials provided 
       with the distribution.
    * Neither the name of the DHUD2 Development Team nor the names of 
       its contributors may be used to endorse or promote products derived 
       from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]

--ACE Version
local VERSION = tonumber(("$Revision 57887$"):match("%d+"))
--Addon Mix
DHUD2 = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "AceDebug-2.0", "FuBarPlugin-2.0")
local DHUD2, self = DHUD2, DHUD2
DHUD2.revision = "r" .. VERSION
DHUD2.versionstring = "0.9b |cffff8888r%d|r"
DHUD2.version = DHUD2.versionstring:format(VERSION)
DHUD2.date = ("$Date$"):match("%d%d%d%d%-%d%d%-%d%d")
local L = AceLibrary("AceLocale-2.2"):new("DHUD2")

local unpack = unpack
local pairs = pairs
local type = type
local _G = _G
local string_find = string.find
local string_format = string.format
local string_len = string.len
local string_sub = string.sub 
local string_gsub = string.gsub
local UnitHealthMax = UnitHealthMax
local UnitHealth = UnitHealth
local UnitInParty = UnitInParty
local UnitIsEnemy = UnitIsEnemy
local UnitManaMax = UnitManaMax
local UnitMana = UnitMana
local UnitIsTapped = UnitIsTapped
local UnitIsTappedByPlayer = UnitIsTappedByPlayer
local UnitPowerType = UnitPowerType
local UnitClassification = UnitClassification
local UnitIsPlayer = UnitIsPlayer
local UnitPlayerControlled = UnitPlayerControlled
local UnitCanAttack = UnitCanAttack
local UnitIsPVP = UnitIsPVP
local UnitCanAttack = UnitCanAttack
local UnitFactionGroup = UnitFactionGroup
local UnitReaction = UnitReaction
local UnitIsPVPFreeForAll = UnitIsPVPFreeForAll
local UnitDebuff = UnitDebuff
local UnitCreatureType = UnitCreatureType
local UnitIsUnit = UnitIsUnit
local UnitClass = UnitClass
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitPlayerOrPetInParty = UnitPlayerOrPetInParty 
local UnitBuff = UnitBuff
local GetPlayerBuff = GetPlayerBuff
local GetPlayerBuffTimeLeft = GetPlayerBuffTimeLeft
local GetRaidTargetIndex = GetRaidTargetIndex
local GetDifficultyColor = GetDifficultyColor
local GetPlayerBuffTimeLeft = GetPlayerBuffTimeLeft
local GetPetHappiness = GetPetHappiness

local targetBorder = ""
local petBorder = ""
local isAlive = false
local isDead = false
local regen = false
local isCasting = false
local inCombat = false
local isDruid = false
local hasPet = false
local regen = false
local currLayout = {
        ["playerHealth"] = e,
        ["playerPower"] = e,
        ["targetHealth"] = e,
        ["targetPower"] = e,
        ["petHealth"] = e,
        ["petPower"] = e,
        ["druidMana"] = e,
        ["playerCastBar"] = e,
        ["playerFlash"] = e,
        ["targetCastBar"] = e,
        ["targetFlash"] = e,
        ["targetTargetHealth"] = e,
        ["targetTargetPower"] = e,
        ["focusHealth"] = e,
        ["focusPower"] = e,
    }

DHUD2.isDruid = isDruid
DHUD2.isAlive = isAlive
DHUD2.isDead = isDead
DHUD2.rBorder = rBorder
DHUD2.lBorder = lBorder
DHUD2.currLayout = currLayout
    
-- Local Functions
local showFrames, myAddons, formatTime, getHealth, getPower
do
    function myAddons()
        if (myAddOnsFrame_Register) then
            local DHUD2_mya = {
                ["name"]         = "DHUD2",
                ["version"]      = versionstring,
                ["author"]       = "Arcanefoam (Horacio Hoyos) -- Drathl/Silberklinge",
                ["category"]     = MYADDONS_CATEGORY_COMBAT,
                ["email"]        = "arcanefoam AT gmail DOT com",
                ["website"]      = "http://www.wowace.com",
            }
            myAddOnsFrame_Register(DHUD2_mya)
        end
    end
    
    --[[
            getHealth: Get the health value for a unit. 
            arguments:
                unit: unit to query value
            returns:
                maximum, current and percent values. nedd: bollean idicating if the unit's health is below 100% --]]
    function getHealth(unit)
        local maxval, curval, curFrac, need, found
        maxval = UnitHealthMax(unit)
        curval = UnitHealth(unit)
        if unit == "player" or unit == "pet" or UnitIsUnit(unit, "pet") then            
            curFrac = curval / maxval
        else
            if UnitInParty(unit) or UnitInRaid(unit) then
                curFrac = curval / maxval
            elseif UnitIsEnemy("player", unit) then
                if MobHealth3 then
                    curval, maxval, found = MobHealth3:GetUnitHealth(unit, curval, 
                    maxval)
                end
                if found then
                    curFrac = curval / maxval
                else
                    curFrac = curval / 100
                end
            else
                curFrac = curval / 100
            end
        end
        if curFrac ~= 1 then 
            need = true
        end
        return maxval, curval, curFrac, need
    end
    
    --[[
            getPower: Get the health value for a unit. 
            arguments:
                unit: unit to query value
                powerType: type of power for that unit
            returns:
                maximum, current and percent values. nedd: bollean idicating if the unit lackcs power (has power if power = rage) --]]
    function getPower(who)
        local maxval, curval, curFrac, need, powerType
        powerType = UnitPowerType(who)
        if powerType then
            maxval = UnitManaMax(who)
            curval = UnitMana(who)
            if maxval ~= 0 then
                curFrac = curval / maxval
                if powerType == 1 then
                    if curFrac ~= 0 then 
                        need = true
                    end
                else
                    if curFrac ~= 1 then 
                        need = true
                    end
                end
            else
                curFrac = 0
                need = false
            end
            return maxval, curval, curFrac, need
        else
            return false
        end
    end

end
DHUD2.getHealth = getHealth
DHUD2.getPower = getPower

 -- General use functions
 --[[
            updateAlpha: Set the HUD's alpha acording to the state of the player
            arguments:
            returns:
                --]]
function DHUD2:updateAlpha()
    local p = self.db.profile
    local alpha
    if isDead then
        alpha = "dead"
    elseif inCombat then
        alpha = "combat"  
    elseif hasTarget then
        alpha = "select"
    elseif regen then
        alpha = "regen"
    elseif isCasting then
        alpha = "regen" 
    else
        alpha = "ooc"
    end
    self:addBackgroundBorder()
    DHUD2_leftFrame:SetAlpha(p.alphas[alpha])
    DHUD2_rightFrame:SetAlpha(p.alphas[alpha])
end

--[[
            rgb2hex: Transform a RGB color to a HEX color
            arguments:
                r, g, b: RGB values
            returns:
                Hex value for the RGB   --]]
function DHUD2:rgb2hex(r, g, b)
    local hexr, hexg, hexb
    hexr = string_format("%x", r * 255)
    hexr = string_len(hexr) == 1 and "0"..hexr or hexr
    hexg = string_format("%x", g * 255)
    hexg = string_len(hexg) == 1 and "0"..hexg or hexg
    hexb = string_format("%x", b * 255)
    hexb = string_len(hexb) == 1 and "0"..hexb or hexb
    return hexr..hexg..hexb
end

--[[
        colorize: Get the bar color according to the mana or power type and percent full. USed for bars and texts
        arguments:
            who: Unit's property being colorized
            propoerty: Property being colorized
            percent: Current percent value for that property
        returns:
            r, g, b and hex values for the calculated color.    --]]
function DHUD2:colorize(who, property, percent)
    local colorTable = self.db.profile.gral.colors
    if not colorTable[who] then return 0,0,0; end
    local r, g, b, diff
    local threshold1, threshold2  = 0.6, 0.2
    local r0, g0, b0 = unpack(colorTable[who][property].color0)
    local r1, g1, b1 = unpack(colorTable[who][property].color1)
    local r2, g2, b2 = unpack(colorTable[who][property].color2)
        
    if ( percent <= threshold2 ) then
        r, g, b = r2, g2, b2
    elseif ( percent <= threshold1) then
        diff = 1 - (percent - threshold2) / (threshold1 - threshold2);
        r = r1 - (r1 - r2) * diff;        
        g = g1 - (g1 - g2) * diff;        
        b = b1 - (b1 - b2) * diff;    
    elseif ( percent < 1) then        
        diff = 1 - (percent - threshold1) / (1 - threshold1);        
        r = r0 - (r0 - r1) * diff;        
        g = g0 - (g0 - g1) * diff;        
        b = b0 - (b0 - b1) * diff;    
    else       
        r, g, b = r0, g0, b0
    end
    if (r < 0) then r = 0; end    
    if (r > 1) then r = 1; end    
    if (g < 0) then g = 0; end    
    if (g > 1) then g = 1; end    
    if (b < 0) then b = 0; end    
    if (b > 1) then b = 1; end
    hex = self:rgb2hex(r, g, b)
    return r, g, b, hex
end

--[[
        checkElite: Check if the unit is elite
        arguments:
            unit: unit to check
        returns:
            icon to use, text to show --]]
function DHUD2:checkElite(unit)
    local el = UnitClassification(unit)
    local eliteValues = {
        ["worldboss"] = {"targetElite", "++"},
        ["elite"] = {"targetElite", "+"},
        ["rareelite"] = {"targetRare", "r+"},
        ["rare"] = {"targetRare", "r"},
        ["normal"] = {false, ""},
        }
    return eliteValues[el][1], eliteValues[el][2]
end

 --[[
        unitIsNPC: Check if the unit is a NPC
        arguments:
            unit: unit to check
        returns:
            true if the unit is a NPC --]]
function DHUD2:unitIsNPC(unit)
    return not (UnitIsPlayer(unit) or UnitPlayerControlled(unit) or UnitCanAttack("player", unit))
end
--[[
        unitIsPet: Check if the unit is a pet
        arguments:
            unit: unit to check
        returns:
            true if the unit is a pet --]]
function self:unitIsPet(unit)
    return not self:unitIsNPC(unit) and not UnitIsPlayer(unit) and UnitPlayerControlled(unit)
end   

local isInit = false
function DHUD2:OnInitialize()
    --self:SetDebugging(true)
    self:Debug("OnInitialize")
    if AceLibrary:HasInstance("Waterfall-1.0") then
        self:Debug("OnInitialize, Waterfall")
		AceLibrary("Waterfall-1.0"):Register('DHUD2',
			'aceOptions', self.options,
			'title', L["DHUD2"],
			'treeLevels', 4,
			'colorR', 0.6, 'colorG', 0.6, 'colorB', 0.8
		)
		self:RegisterChatCommand({"/dhud2"}, function()
			AceLibrary("Waterfall-1.0"):Open('DHUD2')
		end)
		if AceLibrary:HasInstance("Dewdrop-2.0") then
            self:Debug("OnInitialize, Dewdrop")
			self:RegisterChatCommand({"/dhud2dd"}, function()
				AceLibrary("Dewdrop-2.0"):Open('DHDUD2', 'children', function()
					AceLibrary("Dewdrop-2.0"):FeedAceOptionsTable(self.options)
				end)
			end)
		end
		self:RegisterChatCommand({"/dhud2cl"}, self.options)
	elseif AceLibrary:HasInstance("Dewdrop-2.0") then
        self:Debug("OnInitialize, Dewdrop")
		self:RegisterChatCommand({"/dhud2"}, function()
			AceLibrary("Dewdrop-2.0"):Open('DHUD2', 'children', function()
				AceLibrary("Dewdrop-2.0"):FeedAceOptionsTable(self.options)
			end)
		end)
		self:RegisterChatCommand({"/dhud2cl"}, self.options)
	else
        self:Debug("OnInitialize, None")
		self:RegisterChatCommand({"/dhud2"}, self.options)
	end
    self:SetIcon("Interface\\AddOns\\DHUD2\\icon.tga")
    -- MyAddons Support
    myAddons()
    self.isInit = true
end
DHUD2.isInit = isInit

function DHUD2:OnClick()
    if AceLibrary:HasInstance("Waterfall-1.0") then
        AceLibrary("Waterfall-1.0"):Open('DHUD2')
	elseif AceLibrary:HasInstance("Dewdrop-2.0") then
        AceLibrary("Dewdrop-2.0"):Open('DHUD2', 'children', function()
				AceLibrary("Dewdrop-2.0"):FeedAceOptionsTable(self.options)
			end)
	end
end

local isSetup = false
function blizzFrames(self, showPlayer, showTarget, showPet, showCasting)
    if showTarget then
        TargetFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
        TargetFrame:RegisterEvent("UNIT_HEALTH")
        TargetFrame:RegisterEvent("UNIT_LEVEL")
        TargetFrame:RegisterEvent("UNIT_FACTION")
        TargetFrame:RegisterEvent("UNIT_CLASSIFICATION_CHANGED")
        TargetFrame:RegisterEvent("UNIT_AURA")
        TargetFrame:RegisterEvent("PLAYER_FLAGS_CHANGED")
        TargetFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")
        TargetFrame:RegisterEvent("PLAYER_FOCUS_CHANGED")
        if UnitExists("target") then TargetFrame:Show() end
        ComboFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
        ComboFrame:RegisterEvent("PLAYER_COMBO_POINTS")
    else
        TargetFrame:UnregisterEvent("PLAYER_TARGET_CHANGED")
        TargetFrame:UnregisterEvent("UNIT_HEALTH")
        TargetFrame:UnregisterEvent("UNIT_LEVEL")
        TargetFrame:UnregisterEvent("UNIT_FACTION")
        TargetFrame:UnregisterEvent("UNIT_CLASSIFICATION_CHANGED")
        TargetFrame:UnregisterEvent("UNIT_AURA")
        TargetFrame:UnregisterEvent("PLAYER_FLAGS_CHANGED")
        TargetFrame:UnregisterEvent("PARTY_MEMBERS_CHANGED")
        TargetFrame:UnregisterEvent("PLAYER_FOCUS_CHANGED")
        ComboFrame:UnregisterEvent("PLAYER_TARGET_CHANGED")
        ComboFrame:UnregisterEvent("PLAYER_COMBO_POINTS")
        TargetFrame:Hide()
        ComboFrame:Hide()
    end
    
    if showPlayer then
        PlayerFrame:RegisterEvent("UNIT_LEVEL")
        PlayerFrame:RegisterEvent("UNIT_COMBAT")
        PlayerFrame:RegisterEvent("UNIT_SPELLMISS")
        PlayerFrame:RegisterEvent("UNIT_PVP_UPDATE")
        PlayerFrame:RegisterEvent("UNIT_MAXMANA")
        PlayerFrame:RegisterEvent("PLAYER_ENTER_COMBAT")
        PlayerFrame:RegisterEvent("PLAYER_LEAVE_COMBAT")
        PlayerFrame:RegisterEvent("PLAYER_UPDATE_RESTING")
        PlayerFrame:RegisterEvent("PARTY_MEMBERS_CHANGED")
        PlayerFrame:RegisterEvent("PARTY_LEADER_CHANGED")
        PlayerFrame:RegisterEvent("PARTY_LOOT_METHOD_CHANGED")
        PlayerFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
        PlayerFrame:RegisterEvent("PLAYER_REGEN_DISABLED")
        PlayerFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
        PlayerFrameHealthBar:RegisterEvent("UNIT_HEALTH")
        PlayerFrameHealthBar:RegisterEvent("UNIT_MAXHEALTH")
        PlayerFrameManaBar:RegisterEvent("UNIT_MANA")
        PlayerFrameManaBar:RegisterEvent("UNIT_RAGE")
        PlayerFrameManaBar:RegisterEvent("UNIT_FOCUS")
        PlayerFrameManaBar:RegisterEvent("UNIT_ENERGY")
        PlayerFrameManaBar:RegisterEvent("UNIT_HAPPINESS")
        PlayerFrameManaBar:RegisterEvent("UNIT_MAXMANA")
        PlayerFrameManaBar:RegisterEvent("UNIT_MAXRAGE")
        PlayerFrameManaBar:RegisterEvent("UNIT_MAXFOCUS")
        PlayerFrameManaBar:RegisterEvent("UNIT_MAXENERGY")
        PlayerFrameManaBar:RegisterEvent("UNIT_MAXHAPPINESS")
        PlayerFrameManaBar:RegisterEvent("UNIT_DISPLAYPOWER")
        PlayerFrame:RegisterEvent("UNIT_NAME_UPDATE")
        PlayerFrame:RegisterEvent("UNIT_PORTRAIT_UPDATE")
        PlayerFrame:RegisterEvent("UNIT_DISPLAYPOWER")
        PlayerFrame:Show()
    else
        PlayerFrame:UnregisterEvent("UNIT_LEVEL")
        PlayerFrame:UnregisterEvent("UNIT_COMBAT")
        PlayerFrame:UnregisterEvent("UNIT_SPELLMISS")
        PlayerFrame:UnregisterEvent("UNIT_PVP_UPDATE")
        PlayerFrame:UnregisterEvent("UNIT_MAXMANA")
        PlayerFrame:UnregisterEvent("PLAYER_ENTER_COMBAT")
        PlayerFrame:UnregisterEvent("PLAYER_LEAVE_COMBAT")
        PlayerFrame:UnregisterEvent("PLAYER_UPDATE_RESTING")
        PlayerFrame:UnregisterEvent("PARTY_MEMBERS_CHANGED")
        PlayerFrame:UnregisterEvent("PARTY_LEADER_CHANGED")
        PlayerFrame:UnregisterEvent("PARTY_LOOT_METHOD_CHANGED")
        PlayerFrame:UnregisterEvent("PLAYER_ENTERING_WORLD")
        PlayerFrame:UnregisterEvent("PLAYER_REGEN_DISABLED")
        PlayerFrame:UnregisterEvent("PLAYER_REGEN_ENABLED")
        PlayerFrameHealthBar:UnregisterEvent("UNIT_HEALTH")
        PlayerFrameHealthBar:UnregisterEvent("UNIT_MAXHEALTH")
        PlayerFrameManaBar:UnregisterEvent("UNIT_MANA")
        PlayerFrameManaBar:UnregisterEvent("UNIT_RAGE")
        PlayerFrameManaBar:UnregisterEvent("UNIT_FOCUS")
        PlayerFrameManaBar:UnregisterEvent("UNIT_ENERGY")
        PlayerFrameManaBar:UnregisterEvent("UNIT_HAPPINESS")
        PlayerFrameManaBar:UnregisterEvent("UNIT_MAXMANA")
        PlayerFrameManaBar:UnregisterEvent("UNIT_MAXRAGE")
        PlayerFrameManaBar:UnregisterEvent("UNIT_MAXFOCUS")
        PlayerFrameManaBar:UnregisterEvent("UNIT_MAXENERGY")
        PlayerFrameManaBar:UnregisterEvent("UNIT_MAXHAPPINESS")
        PlayerFrameManaBar:UnregisterEvent("UNIT_DISPLAYPOWER")
        PlayerFrame:UnregisterEvent("UNIT_NAME_UPDATE")
        PlayerFrame:UnregisterEvent("UNIT_PORTRAIT_UPDATE")
        PlayerFrame:UnregisterEvent("UNIT_DISPLAYPOWER")
        PlayerFrame:Hide()
    end    
    
    if showCasting then
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_START");
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_STOP");
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_FAILED");
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED");
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_DELAYED");
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START");
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE");
        CastingBarFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP");
    else
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_START")
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_STOP")
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_FAILED")
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED")        
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_DELAYED")
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START")
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
        CastingBarFrame:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
        CastingBarFrame:Hide()
    end
    
    if showPet then
        
    end
    
end
DHUD2.isSetup = isSetup
DHUD2.blizzFrames = blizzFrames


-- Asing each bar the their respective tracking stat accordingly to the saved profile
function asignBars()
    local l = self.db.profile.layout
    -- Player Bars
    self:Debug("playerBars: ",l.playerBars )
    if l.playerBars == "HL~PR" then
        currLayout.playerHealth = "li"
        currLayout.playerPower = "ri"
        currLayout.targetHealth = "lm"
        currLayout.targetPower = "rm"
        currLayout.playerCastBar = "rc"
        currLayout.playerFlash = "rf"
        currLayout.targetCastBar = "lc"
        currLayout.targetFlash = "lf"        
    elseif l.playerBars == "HR~PL" then
        currLayout.playerHealth = "ri"
        currLayout.playerPower = "li"
        currLayout.targetHealth = "rm"
        currLayout.targetPower = "lm"
        currLayout.playerCastBar = "lc"
        currLayout.playerFlash = "lf"
        currLayout.targetCastBar = "rc"
        currLayout.targetFlash = "rf"  
    elseif l.playerBars == "HI~PM" then
        if l.player == L["Left"] then
            currLayout.playerHealth = "li"
            currLayout.playerPower = "lm"
            currLayout.playerCastBar = "lc"
            currLayout.playerFlash = "lf"
            currLayout.targetHealth = "ri"
            currLayout.targetPower = "rm"
            currLayout.targetCastBar = "rc"
            currLayout.targetFlash = "rf"
        else
            currLayout.playerHealth = "ri"
            currLayout.playerPower = "rm"
            currLayout.playerCastBar = "rc"
            currLayout.playerFlash = "rf"
            currLayout.targetHealth = "li"
            currLayout.targetPower = "lm"
            currLayout.targetCastBar = "lc"
            currLayout.targetFlash = "lf"
        end
    elseif l.playerBars == "HO~PI" then
        if l.player == L["Left"] then
            currLayout.playerHealth = "lm"
            currLayout.playerPower = "li"
            currLayout.playerCastBar = "lc"
            currLayout.playerFlash = "lf"
            currLayout.targetHealth = "rm"
            currLayout.targetPower = "ri"
            currLayout.targetCastBar = "rc"
            currLayout.targetFlash = "rf"
        else
            currLayout.playerHealth = "rm"
            currLayout.playerPower = "ri"
            currLayout.playerCastBar = "rc"
            currLayout.playerFlash = "rf"
            currLayout.targetHealth = "lm"
            currLayout.targetPower = "li"
            currLayout.targetCastBar = "lc"
            currLayout.targetFlash = "lf"
        end
    end
    
    -- Pet bars
    self:Debug("petBars: ",l.petBars )
    if l.petBars == "HL~PR" then
        currLayout.petHealth = "lpo"
        currLayout.petPower = "rpo"
    elseif l.petBars == "HI~PO" then
        if l.pet == L["Left"] then
            currLayout.petHealth = "lpi"
            currLayout.petPower = "lpo"
        else
            currLayout.petHealth = "rpi"
            currLayout.petPower = "rpo"
        end
    elseif l.petBars == "HO~PI" then
        if l.pet == L["Left"] then
            currLayout.petHealth = "lpo"
            currLayout.petPower = "lpi"
        else
            currLayout.petHealth = "rpo"
            currLayout.petPower = "rpi"
        end
    end
    
    -- Druid bar
    if l.druidBar == "R" then
        currLayout.druidMana = "rpo"
    else
        currLayout.druidMana = "lpo"
    end
    
    -- Outer bars
    local outer = self.db.profile.outer
    if outer.right == "TTH" then
        currLayout.targetTargetHealth = "ro"
    elseif outer.left == "TTH" then
        currLayout.targetTargetHealth = "lo"
    else
        currLayout.targetTargetHealth = "e"
    end
    if outer.right == "TTP" then
        currLayout.targetTargetPower = "ro"
    elseif outer.left == "TTP" then
        currLayout.targetTargetPower = "lo"
    else
        currLayout.targetTargetPower = "e"
    end    
    if outer.left == "FH" then
        currLayout.focusHealth = "lo"
    elseif outer.right == "FH" then
        currLayout.focusHealth = "ro"
    else
        currLayout.focusHealth = "e"
    end
    if outer.left == "FP" then
        currLayout.focusPower = "lo"
    elseif outer.right == "FP" then
        currLayout.focusPower = "ro"
    else
        currLayout.focusPower = "e"
    end
    
    for k, v in pairs(currLayout) do
        self:Debug("currLayout: ", k, v)
    end
end

-- Initializa the respective bars with their values and hide the unused ones
function DHUD2:setup()
    self:Debug("setup")
    local maxVal, curVal, curFrac
    local color = {r=0,g=0,b=0,hex=""}
    local p = self.db.profile
    
    -- Set the HUD scale
    DHUD2_mainFrame:SetScale(p.gral.scale)
    
    -- Designate de bars accordingly to the profile
    asignBars()
    self:initBorders()
    local _, class = UnitClass("player")
    DHUD2_lpi:Hide()
    DHUD2_lpo:Hide()
    DHUD2_rpi:Hide()
    DHUD2_rpo:Hide()
    if class == "DRUID" then
        -- Init druid bar
        isDruid = true
        if p.show.druidBar and SmartyCat then 
            self:RegisterEvent("UNIT_DISPLAYPOWER")
            if UnitPowerType("player") ~= 0 then                
                _G["DHUD2_"..currLayout.druidMana]:Show()
                self:RegisterEvent("SmartyCat_ManaUpdated", "OnFormUpdateMana")
                curVal, maxVal = SmartyCat.GetCurrMana(), SmartyCat.GetMaxMana()
                color.r, color.g, color.b, color.hex = self:colorize("player", 0,  curVal/maxVal)
                self:updateBar(currLayout.druidMana, maxVal, curVal, curVal/maxVal, color)
            end
        end
    else
        --Init Pet frames
        hasPet = UnitExists("pet")
        self:Debug("Init pet bars")
        if hasPet and p.show.pet then
            happ = GetPetHappiness()
            maxVal, curVal, curFrac, need = getHealth("pet")
            color.r, color.g, color.b, color.hex = self:colorize("pet", "health",  curFrac)
            self:updateBar(currLayout.petHealth, maxVal, curVal, curFrac, color)
            self:updatePetIcon(petCurH, happ)
            maxVal, curVal, curFrac, need = getPower("pet")
            color.r, color.g, color.b, color.hex = self:colorize("pet", UnitPowerType("pet"),  curFrac)
            self:updateBar(currLayout.petPower, maxVal, curVal, curFrac, color)
        else
            self:updatePetIcon(0, false)
            self:updateBar(currLayout.petHealth, 0, 0, 0)
            self:updateBar(currLayout.petPower, 0, 0, 0)
        end
    end
    
    -- Init Player Bars
    self:Debug("Init player bars")
    isAlive = not UnitIsDeadOrGhost("player")
    isDead = not isAlive
    maxVal, curVal, curFrac, need = getHealth("player")
    color.r, color.g, color.b, color.hex = self:colorize("player", "health",  curFrac)
    self:updateBar(currLayout.playerHealth, maxVal, curVal, curFrac, color)
    maxVal, curVal, curFrac, playerNeedP = getPower("player")
    color.r, color.g, color.b, color.hex = self:colorize("player", UnitPowerType("player"),  curFrac)
    self:updateBar(currLayout.playerPower, maxVal, curVal, curFrac, color)
    self:updateCombos()
    self:updatePlayerPvP()
    if IsResting() and p.show.restIcon then
        DHUD2_playerResting:Show()
    else
        DHUD2_playerResting:Hide()
    end
    if UnitPlayerOrPetInParty("player") then
        if IsPartyLeader() then
            DHUD2_playerLeader:Show()
            local lootmethod = GetLootMethod()
            if lootmethod == "master" then
                DHUD2_playerLooter:Show()
            else
                DHUD2_playerLooter:Hide()
            end
        else
            DHUD2_playerLeader:Hide()
            DHUD2_playerLooter:Hide()
        end
    else
        DHUD2_playerLeader:Hide()
        DHUD2_playerLooter:Hide()
    end        
    DHUD2_playerInCombat:Hide()
    
    -- Init Target frames
    self:Debug("Init target bars")
    if UnitExists("target") and p.show.target then
        maxVal, curVal, curFrac = getHealth("target")
        color.r, color.g, color.b, color.hex = self:colorize("target", "health",  curFrac)
        self:updateBar(currLayout.targetHealth, maxVal, curVal, curFrac, color)
        maxVal, curVal, curFrac = getPower("target")
        color.r, color.g, color.b, color.hex = self:colorize("target", UnitPowerType("target"),  curFrac)
        self:updateBar(currLayout.targetPower, maxVal, curVal, curFrac, color)
        if self:checkElite("target") and p.show.eliteIcon then
            local images = self.layout.images
            local _, _, _, _, paths = unpack(images.left)
            local texture = paths[elite]
            DHUD2_targetElite_Texture:SetTexture(texture)
            DHUD2_targetElite:Show()
        else
            DHUD2_targetElite:Hide()
        end
        self:updateText("target", p.show.target)
        self:updateTargetAuras(true)
    else
        self:updateBar(currLayout.targetHealth, 0, 0, 0)      
        self:updateBar(currLayout.targetPower, 0, 0, 0)      
        self:updateTargetAuras(false)
        DHUD2_targetElite:Hide()
        self:updateText("target", false)
    end
    -- Init Target Target
    self:updateText("targetTarget", UnitExists("targetTarget") and p.show.targetTargetName)
    
    -- Outer Bars
    DHUD2_lo:Hide()
    DHUD2_ro:Hide()
    if UnitExists("targettarget") then
        if currLayout.targetTargetHealth ~= "e" then
            maxVal, curVal, curFrac = getHealth("targettarget")
            color.r, color.g, color.b, color.hex = self:colorize("targetTarget", "health",  curFrac)
            self:updateBar(currLayout.targetTargetHealth, maxVal, curVal, curFrac, color)            
        end
        if currLayout.targetTargetPower ~= "e" then
            maxVal, curVal, curFrac = getPower("targettarget")
            color.r, color.g, color.b, color.hex = self:colorize("targetTarget", UnitPowerType("targetarget"),  curFrac)
            self:updateBar(currLayout.targetTargetPower, maxVal, curVal, curFrac, color)            
        end
    end
    if UnitExists("focus") then
        if currLayout.focusHealth ~= "e" then
            maxVal, curVal, curFrac = getHealth("focus")
            color.r, color.g, color.b, color.hex = self:colorize("focus", "health",  curFrac)
            self:updateBar(currLayout.focusHealth, maxVal, curVal, curFrac, color)            
        end
        if currLayout.focusPower ~= "e" then
            maxVal, curVal, curFrac = getPower("focus")
            color.r, color.g, color.b, color.hex = self:colorize("focus", UnitPowerType("focus"), curFrac)
            self:updateBar(currLayout.focusPower, maxVal, curVal, curFrac, color)            
        end
    end
 
    -- Cast Bars
    self:Debug("Init Cast Bars")
    DHUD2_lc:Hide()
    DHUD2_rc:Hide()
    DHUD2_lf:Hide()
    DHUD2_rf:Hide()
    
    --Set fonts
    self:setFonts()
    
    -- Manage Blizz Frames 
    blizzFrames(self, p.show.blizPlayer, p.show.blizTarget, p.show.blizPet, p.show.blizCasting)
    
    self:updateAlpha()
    self.isSetup = true
end

local createFrames
do
    function createFrames()
        -- Get layout information
        local mainFrame = self.layout.mainFrame
        local subFrames = self.layout.frames
        local icons = self.layout.icons
        local bars = self.layout.bars
        local barText = self.layout.barText
        local texts = self.layout.texts
        local tRA = self.layout.targetRightAuras
        local tLA = self.layout.targetLeftAuras
        local playBuff = self.layout.playbuff
      
        -- Main Frame
        local point, parent, relative, x, y, width, height = unpack(mainFrame)
        ref = CreateFrame ("Frame", "DHUD2_mainFrame", _G[parent])
        ref:ClearAllPoints()
        ref:SetPoint(point, parent, relative, x, y)
        ref:SetHeight(height)
        ref:SetWidth(width)
        ref:EnableMouse(false)
        ref:Show()
        
        -- Left and right frames were the bars and texts are placed
        for k, v in pairs(subFrames) do
            point, parent, relative, x, y, width, height, x0, x1, y0, y1, path = unpack(v)
            ref = CreateFrame("Frame", "DHUD2_"..k, _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            ref:SetHeight(height)
            ref:SetWidth(width)
            local bgt = ref:CreateTexture("DHUD2_"..k.."_Texture", "BACKGROUND")
            bgt:SetTexture(path)
            bgt:ClearAllPoints()
            bgt:SetPoint("TOPLEFT", ref , "TOPLEFT", 0, 0)
            bgt:SetPoint("BOTTOMRIGHT", ref , "BOTTOMRIGHT", 0, 0)
            bgt:SetTexCoord(x0, x1, y0, y1)
            ref:EnableMouse(false)
            --ref:Hide()
        end
        
        -- Bars
        for k, v in pairs(bars) do
            point, parent, relative, x, y, width, height, x0, x1, y0, y1, path = unpack(v)
            ref = CreateFrame("Frame", "DHUD2_"..k, _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent , relative, x, y)
            ref:SetHeight(height)
            ref:SetWidth(width)
            bgt = ref:CreateTexture("DHUD2_"..k.."_Texture", "BACKGROUND")
            bgt:SetTexture(path)
            bgt:SetPoint(point, ref, relative, 0, 0)
            bgt:SetHeight(height)
            bgt:SetWidth(width)
            bgt:SetTexCoord(x0, x1, y0, y1)
            ref:EnableMouse(false)
            ref:Hide()
        end

        -- Bar's Text
        for k, v in pairs(barText) do
            point, parent, relative, x, y, width, height, justify = unpack(v)
            ref = CreateFrame("Button", "DHUD2_"..k.."_Text", _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            local font = ref:CreateFontString("DHUD2_"..k.."_Text_Text", "OVERLAY")
            font:SetFontObject(GameFontHighlightSmall)
            font:SetText(k.."30000")
            font:SetWidth(font:GetStringWidth())
            font:SetHeight(height)
            font:ClearAllPoints()
            font:SetPoint(point, ref, relative, 0, 0)
            font:SetJustifyH(justify)
            font:Show()
            ref:SetHeight(height)
            ref:SetWidth(font:GetStringWidth())
            ref:Hide()
        end

        -- Text
        for k, v in pairs(texts) do
            point, parent, relative, x, y, width, height = unpack(v)
            ref = CreateFrame("Button", "DHUD2_"..k.."_Text", _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            local font = ref:CreateFontString("DHUD2_"..k.."_Text_Text", "ARTWORK")
            font:SetFontObject(GameFontHighlightSmall)
            font:SetText(k)
            font:SetJustifyH("CENTER")
            font:SetWidth(font:GetStringWidth())
            font:SetHeight(height)
            font:Show()
            font:ClearAllPoints()
            font:SetPoint(point, ref, relative, 0, 0)
            ref:SetHeight(height)
            ref:SetWidth(font:GetStringWidth())
            ref:Hide()
        end
        
        -- Icons
        for k, v in pairs(icons) do
            point, parent, relative, x, y, width, height, x0, x1, y0, y1, path = unpack(v)
            if string_find(k, "PvP") then         
                local faction = UnitFactionGroup("player")
                if faction == "Horde" then
                    path = path[2]
                else
                    path = path[1]
                end
            end
            ref = CreateFrame("Frame", "DHUD2_"..k, _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            ref:SetHeight(height)
            ref:SetWidth(width)
            local bgt = ref:CreateTexture("DHUD2_"..k.."_Texture","BACKGROUND")
            bgt:SetTexture(path)
            bgt:ClearAllPoints()
            bgt:SetPoint("TOPLEFT", ref , "TOPLEFT", 0, 0)
            bgt:SetPoint("BOTTOMRIGHT", ref , "BOTTOMRIGHT", 0, 0)
            bgt:SetTexCoord(x0,x1,y0,y1)
            ref:Hide()
        end
        
        -- Target Auras
        for i = 1, 40 do
            point, parent, relative, x, y, width, height = unpack(tRA["TRA"..i])
            ref = CreateFrame("Button", "DHUD2_TRA"..i, _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            ref:SetHeight(height)
            ref:SetWidth(width)
            local font = ref:CreateFontString("DHUD2_TRA"..i.."_Text", "ARTWORK")
            font:SetFontObject(GameFontHighlightSmall)
            font:SetText("")
            font:SetJustifyH("RIGHT")
            font:SetJustifyV("BOTTOM")
            font:SetWidth(width+1)
            font:SetHeight(height-5)
            font:SetFont("FRIZQT__.TTF", width / 2.2, "OUTLINE")
            font:Show()
            font:ClearAllPoints()
            font:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            bgt = ref:CreateTexture("DHUD2_TRA"..i.."_Border", "OVERLAY")
            bgt:SetTexture("Interface\\Buttons\\UI-Debuff-Border")
            bgt:SetPoint("BOTTOM", ref, "BOTTOM", 0, 0)
            bgt:SetHeight(height)
            bgt:SetWidth(width)
            bgt:SetTexCoord(0, 1, 0, 1)
            ref:SetNormalTexture( "Interface\\Icons\\Ability_Druid_TravelForm" )
            ref:SetScript("OnEnter", function(this)
                    if not (this:IsVisible() and self.db.profile.show.auraTips) then
                        return
                    end
                    GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
                    if this.hasdebuff then
                        GameTooltip:SetUnitDebuff(this.unit, this.id)
                    else
                        GameTooltip:SetUnitBuff(this.unit, this.id)
                    end
                end )
            ref:SetScript("OnLeave", function() 
                    GameTooltip:Hide()
                end )        
            ref:EnableMouse(true)
            ref:Hide()
        end
        
        for i = 1, 40 do
            point, parent, relative, x, y, width, height = unpack(tLA["TLA"..i])
            ref = CreateFrame("Button", "DHUD2_TLA"..i, _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            ref:SetHeight(height)
            ref:SetWidth(width)
            local font = ref:CreateFontString("DHUD2_TLA"..i.."_Text", "ARTWORK")
            font:SetFontObject(GameFontHighlightSmall)
            font:SetText("")
            font:SetJustifyH("RIGHT")
            font:SetJustifyV("BOTTOM")
            font:SetWidth(width+1)
            font:SetHeight(height-5)
            font:SetFont("FRIZQT__.TTF", width / 2.2, "OUTLINE")
            font:Show()
            font:ClearAllPoints()
            font:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            bgt = ref:CreateTexture("DHUD2_TLA"..i.."_Border","OVERLAY")
            bgt:SetTexture("Interface\\Buttons\\UI-Debuff-Border")
            bgt:SetPoint("BOTTOM", ref, "BOTTOM", 0, 0)
            bgt:SetHeight(height)
            bgt:SetWidth(width)
            bgt:SetTexCoord(0, 1, 0, 1)
            ref:SetNormalTexture( "Interface\\Icons\\Ability_Druid_TravelForm" )
            ref:SetScript("OnEnter", function(this)
                    if not (this:IsVisible() and self.db.profile.show.auraTips) then
                        return
                    end
                    GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
                    if this.hasdebuff then
                        GameTooltip:SetUnitDebuff(this.unit, this.id)
                    else
                        GameTooltip:SetUnitBuff(this.unit, this.id)
                    end
                end )
            ref:SetScript("OnLeave", function() 
                    GameTooltip:Hide()
                end )        
            ref:EnableMouse(true)
            ref:Hide()
        end
        
        -- Player Buff
        for k, v in pairs(playBuff) do
            point, parent, relative, x, y, width, height = unpack(v)
            ref = CreateFrame("Button", "DHUD2_"..k, _G["DHUD2_"..parent])
            ref:ClearAllPoints()
            ref:SetPoint(point, "DHUD2_"..parent, relative, x, y)
            ref:SetHeight(height)
            ref:SetWidth(width)
            font = ref:CreateFontString("DHUD2_"..k.."_Text", "OVERLAY")
            font:ClearAllPoints()
            font:SetFontObject(GameFontHighlightSmall)
            font:SetPoint("CENTER", ref, "CENTER", 0,0)
            font:SetText("")
            font:SetJustifyH("CENTER")
            font:SetJustifyV("CENTER")
            font:SetWidth(width*2)
            font:SetHeight(height)
            bgt = ref:CreateTexture("DHUD2_"..k.."_Border", "OVERLAY")
            bgt:SetTexture("Interface\\AddOns\\DHUD2\\textures\\serenity0")
            bgt:SetVertexColor(1.0, 1.0, 1.0)
            bgt:SetPoint("CENTER", ref, "CENTER", 0, 0)
            bgt:SetHeight(height*1.7)
            bgt:SetWidth(width*1.7)
            bgt:SetTexCoord(0, 1, 0, 1)
            bgt:SetBlendMode("BLEND")                
            ref:SetNormalTexture( "Interface\\AddOns\\DHUD2\\icon" )
            ref:SetScript("OnEnter", function(this) 
                    if not (this:IsVisible() and self.db.profile.show.auraTips) or (this:GetEffectiveAlpha() == 0) then
                        return
                    end
                    GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT")
                    GameTooltip:SetPlayerBuff(this.id)
                end )
            ref:SetScript("OnLeave", function() 
                    GameTooltip:Hide()
                end )         
            ref:Hide()
        end
    end
end

local addonLoaded = false
function DHUD2:OnEnable() -- ok
    self:Debug("OnEnable")
    createFrames()
    self:RegisterEvent("PLAYER_ENTERING_WORLD")
    self:RegisterEvent("PLAYER_TARGET_CHANGED")
    self:RegisterEvent("PLAYER_REGEN_ENABLED")
    self:RegisterEvent("PLAYER_REGEN_DISABLED")
    self:RegisterEvent("PLAYER_AURAS_CHANGED")
    self:RegisterEvent("PLAYER_COMBO_POINTS")
    self:RegisterEvent("PLAYER_ALIVE")
    self:RegisterEvent("PLAYER_UNGHOST")
    self:RegisterEvent("PLAYER_DEAD")
    self:RegisterEvent("PLAYER_UPDATE_RESTING")
    self:RegisterEvent("PLAYER_FOCUS_CHANGED")
    self:RegisterEvent("UNIT_HEALTH", "onUnitHealth")
    self:RegisterEvent("UNIT_MAXHEALTH", "onUnitHealth")
    self:RegisterEvent("UNIT_MANA", "onUnitPower")
    self:RegisterEvent("UNIT_MAXMANA", "onUnitPower")  
    self:RegisterEvent("UNIT_FOCUS", "onUnitPower")
    self:RegisterEvent("UNIT_MAXFOCUS", "onUnitPower")
    self:RegisterEvent("UNIT_RAGE", "onUnitPower")
    self:RegisterEvent("UNIT_MAXRAGE", "onUnitPower")
    self:RegisterEvent("UNIT_ENERGY", "onUnitPower")
    self:RegisterEvent("UNIT_MAXENERGY", "onUnitPower")
    self:RegisterEvent("UNIT_FACTION")
    self:RegisterEvent("UNIT_AURA")
    self:RegisterEvent("UNIT_TARGET")
    self:RegisterEvent("UNIT_HAPPINESS")
    self:RegisterEvent("UNIT_PET")
    self:RegisterEvent("UNIT_DISPLAYPOWER")
    self:RegisterEvent("RAID_TARGET_UPDATE")
    self:RegisterEvent("PARTY_LOOT_METHOD_CHANGED")
    self:RegisterEvent("PARTY_MEMBERS_CHANGED")

     -- Target Dropdown
    DHUD2_target_Text:RegisterForClicks('RightButtonUp')
    DHUD2_target_Text:SetScript("OnClick", function()
        self:Debug("target_Text->OnClick")
        ToggleDropDownMenu(1, nil, DHUD2_targetDropDown, "DHUD2_target_Text", 25, 10)
    end )
    --[[
    if p.show.casting then
        castTimer = 0
        self:RegisterEvent("UNIT_SPELLCAST_START")
        self:RegisterEvent("UNIT_SPELLCAST_STOP")
        self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
        self:RegisterEvent("UNIT_SPELLCAST_FAILED", "OnSpellFail")
        self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED", "OnSpellFail")
        self:RegisterEvent("UNIT_SPELLCAST_DELAYED")
        self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
        self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_UPDATE")
    end
    ]]
    addonLoaded = true
end

--Reset the init variables so the addon will reload correctly when re-enabled
function DHUD2:OnDisable()
    addonLoaded = false
    isSetup = false
    isInit = false
end

-- Load de configuration menu/window libraries and do initial setup
function DHUD2:PLAYER_ENTERING_WORLD()
    self:Debug("PLAYER_ENTERING_WORLD")
    if not self.isInit then
        self:OnInitialize()
    end
    if self.isSetup then
        return
    end
    self:setup()
end

-- Update Target and TargetTarget
function DHUD2:PLAYER_TARGET_CHANGED()
    self:Debug("PLAYER_TARGET_CHANGED")
    local maxVal, curVal, curFrac
    local color = {r=0,g=0,b=0,hex=""}
    local p = self.db.profile
    if UnitExists("target") and p.show.target and not(self:unitIsNPC("target") and not p.show.npc) then
        hasTarget = true
        maxVal, curVal, curFrac = getHealth("target")
        color.r, color.g, color.b, color.hex = self:colorize("target", "health", curVal)
        self:updateBar(currLayout.targetHealth, maxVal, curVal, curFrac, color)
        maxVal, curVal, curFrac, _ = getPower("target")
        color.r, color.g, color.b, color.hex = self:colorize("target", UnitPowerType("target"), curVal)
        self:updateBar(currLayout.targetPower, maxVal, curVal, curFrac, color)
        self:updateCombos()
        self:updateText("target", p.show.target)
        -- make name clickable?
        if ( UnitIsUnit("target", "player") and ( GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0 )) or
           ( UnitIsPlayer("target")  and not UnitIsEnemy("player", "target")  and not UnitIsUnit("target", "player") ) or
           UnitIsUnit("target", "pet") then
            DHUD2_target_Text:EnableMouse(true)
        else
            DHUD2_target_Text:EnableMouse(false)
        end
    else
        hasTarget = false
        self:updateCombos()
        self:updateBar(currLayout.targetHealth, 0, 0, 0)
        self:updateBar(currLayout.targetPower, 0, 0, 0)
        DHUD2_target_Text:Hide()
        DHUD2_targetTarget_Text:Hide()
    end
    self:updateTargetIcon(hasTarget)
    self:updateTargetAuras(hasTarget)
    --Target target
    if UnitExists("targettarget") and not(self:unitIsNPC("targettarget") and not p.show.npc) and not UnitIsUnit("targetarget", "player") then
        if currLayout.targetTargetHealth ~= "e" then
            maxVal, curVal, curFrac = getHealth("targettarget")
            color.r, color.g, color.b, color.hex = self:colorize("targetTarget", "health", curVal)
            self:updateBar(currLayout.targetTargetHealth, maxVal, curVal, curFrac, color)
        end
        if currLayout.targetTargetPower ~= "e" then
            maxVal, curVal, curFrac, _ = getPower("targettarget")
            color.r, color.g, color.b, color.hex = self:colorize("targetTarget", UnitPowerType("targettarget"), curVal)
            self:updateBar(currLayout.targetTargetPower, maxVal, curVal, curFrac, color)
        end
        self:updateText("targetTarget", p.show.targetTargetName)
    else
        if currLayout.targetTargetHealth ~= "e" then
            self:updateBar(currLayout.targetTargetHealth, 0, 0, 0)
        end
        if currLayout.targetTargetPower ~= "e" then
            self:updateBar(currLayout.targetTargetPower, 0, 0, 0)
        end
        self:updateText("targetTarget", false)
    end
    self:updateAlpha()
end

function DHUD2:PLAYER_REGEN_ENABLED()
    self:Debug("PLAYER_REGEN_ENABLED")
    regen = true
    inCombat = false
    DHUD2_playerInCombat:Hide()
    self:updateAlpha()
end

function DHUD2:PLAYER_REGEN_DISABLED()
    self:Debug("PLAYER_REGEN_DISABLED")
    regen = false
    inCombat = true
    if self.db.profile.show.combatIcon then
        DHUD2_playerInCombat:Show()
    end
    self:updateAlpha()
end  

function DHUD2:PLAYER_AURAS_CHANGED()
    self:Debug("PLAYER_AURAS_CHANGED")
    self:updatePlayerAuras()
    self:updateAlpha()
end

function DHUD2:PLAYER_COMBO_POINTS()
    self:Debug("PLAYER_COMBO_POINTS")
    self:updateCombos()
end

function DHUD2:PLAYER_ALIVE()
    self:Debug("PLAYER_ALIVE")
    if UnitIsDeadOrGhost("player") then
        self:PLAYER_DEAD()
    else
        self:OnPlayerLife()
    end
end

function DHUD2:OnPlayerLife()
    self:Debug("OnPlayerLife")
    isAlive = true
    isDead = false
    self:updateAlpha()
end

function DHUD2:PLAYER_UNGHOST()
    self:Debug("PLAYER_UNGHOST")
    isAlive = true
    isDead = false
    self:updateAlpha()
end

function DHUD2:PLAYER_DEAD()
    self:Debug("PLAYER_DEAD")
    inCombat = false
    isDead = true
    isAlive = false
    playerNeedH = false
    playerNeedP = false
    petNeedH = false
    petNeedP = false
    isCasting = false
    self:updateAlpha()
end  

function DHUD2:PLAYER_UPDATE_RESTING()
    self:Debug("PLAYER_UPDATE_RESTING")
    if IsResting() and self.db.profile.show.restIcon and isAlive then
        DHUD2_playerResting:Show();
    else
        DHUD2_playerResting:Hide();
    end
end

function DHUD2:PLAYER_FOCUS_CHANGED()
    self:Debug("PLAYER_FOCUS_CHANGED")
    if UnitExists("focus") then
        local color = {r=0,g=0,b=0,hex=""}
        local maxVal, curVal, curFrac
        if currLayout.focusHealth ~= "e" then
            maxVal, curVal, curFrac = getHealth("focus")
            color.r, color.g, color.b, color.hex = self:colorize("focus", "health",  curFrac)
            self:updateBar(currLayout.focusHealth, maxVal, curVal, curFrac, color) 
        end
        if currLayout.focusPower ~= "e" then
            maxVal, curVal, curFrac = getPower("focus")
            color.r, color.g, color.b, color.hex = self:colorize("focus", UnitPowerType("focus"), curFrac)
            self:updateBar(currLayout.focusPower, maxVal, curVal, curFrac, color)
        end
    else
        if currLayout.focusHealth ~= "e" then
            self:updateBar(currLayout.focusHealth, 0, 0, 0)            
        end
        if currLayout.focusPower ~= "e" then
            self:updateBar(currLayout.focusPower, 0, 0, 0)            
        end
    end
    self:updateAlpha()
end

function DHUD2:onUnitHealth(who)
    self:Debug("onUnitHealth", who)
    local maxVal, curVal, curFrac
    local p = self.db.profile
    local color = {r=0,g=0,b=0,hex=""}
    local typ = "health"
    if who == "player" then
        maxVal, curVal, curFrac, regen = getHealth("player")
        color.r, color.g, color.b, color.hex = self:colorize("player", typ, curVal)
        self:updateBar(currLayout.playerHealth, maxVal, curVal, curFrac, color)
    elseif who == "target" and p.show.target then
        maxVal, curVal, curFrac = getHealth("target")
        if (UnitIsTapped("target") and (not UnitIsTappedByPlayer("target"))) then               
              typ = "tapped"
        end
        color.r, color.g, color.b, color.hex = self:colorize("target", typ, curVal)
        self:updateBar(currLayout.targetHealth, maxVal, curVal, curFrac, color)
    elseif who == "pet" and hasPet and p.show.pet then
        maxVal, curVal, curFrac, regen = getHealth("pet")
        color.r, color.g, color.b, color.hex = self:colorize("pet", typ, curVal)
        self:updateBar(currLayout.petHealth, maxVal, curVal, curFrac, color)
    elseif who == "targettarget" then
        if currLayout.targetTargetHealth ~= "e" then
            maxVal, curVal, curFrac = getHealth("targettarget")
            if (UnitIsTapped("targettarget") and (not UnitIsTappedByPlayer("targettarget"))) then               
                  typ = "tapped"
            end
            color.r, color.g, color.b, color.hex = self:colorize("targetTarget", typ,  curFrac)
            self:updateBar(currLayout.targetTargetHealth, maxVal, curVal, curFrac, color)            
        end
    elseif who == "focus" then
        if currLayout.focusHealth ~= "e" then
            maxVal, curVal, curFrac = getHealth("focus")
            if (UnitIsTapped("focus") and (not UnitIsTappedByPlayer("focus"))) then               
                  typ = "tapped"
            end
            color.r, color.g, color.b, color.hex = self:colorize("focus", typ,  curFrac)
            self:updateBar(currLayout.focusHealth, maxVal, curVal, curFrac, color)            
        end

    end
    self:updateAlpha()
end

function DHUD2:onUnitPower(who)
    self:Debug("onUnitPower", who)
    local maxVal, curVal, curFrac
    local p = self.db.profile
    local color = {r=0,g=0,b=0,hex=""}
    local typ
    typ = UnitPowerType(who)
    if who == "player" then
        maxVal, curVal, curFrac, regen = getPower("player")
        color.r, color.g, color.b, color.hex = self:colorize("player", typ, curFrac)
        self:updateBar(currLayout.playerPower, maxVal, curVal, curFrac, color)
    elseif who == "target" and p.show.target then
        maxVal, curVal, curFrac = getPower("target")
        color.r, color.g, color.b, color.hex = self:colorize("target", typ, curFrac)
        self:updateBar(currLayout.targetPower, maxVal, curVal, curFrac, color)
    elseif who == "pet" and hasPet and p.show.pet then
        maxVal, curVal, curFrac, regen = getPower("pet")
        color.r, color.g, color.b, color.hex = self:colorize("pet", typ, curFrac)
        self:updateBar(currLayout.petPower, maxVal, curVal, curFrac, color)
    elseif who == "targettarget" then    
        if currLayout.targetTargetPower ~= "e" then
            maxVal, curVal, curFrac = getPower("targettarget")
            color.r, color.g, color.b, color.hex = self:colorize("targetTarget", typ,  curFrac)
            self:updateBar(currLayout.targetTargetPower, maxVal, curVal, curFrac, color)            
        end
    elseif who == "focus" then
        if currLayout.focusPower ~= "e" then
            maxVal, curVal, curFrac = getPower("focus")
            color.r, color.g, color.b, color.hex = self:colorize("focus", typ, curFrac)
            self:updateBar(currLayout.focusPower, maxVal, curVal, curFrac, color)            
        end
    end
    self:updateAlpha()
end

function DHUD2:UNIT_FACTION(who)
    local p = self.db.profile
    if who == "player" then
        self:updatePlayerPvP()
        if UnitIsUnit("player", "target") then
            self:updateTargetIcon()
        end
    elseif who == "target" and p.show.target then
        self:updateTargetIcon()
    end
end

function DHUD2:UNIT_AURA(who)
    if who == "target" and self.db.profile.show.target then
        self:updateTargetAuras(true)
    end
end

function DHUD2:UNIT_TARGET(who)
    self:Debug("UNIT_TARGET", who)
    local p = self.db.profile
    if who == "target" then
        if UnitExists("targetTarget") then
            local color = {r=0,g=0,b=0,hex=""}
            local maxVal, curVal, curFrac
            if currLayout.targetTargetHealth ~= "e" then
                maxVal, curVal, curFrac = getHealth("targettarget")
                color.r, color.g, color.b, color.hex = self:colorize("targetTarget", "health",  curFrac)
                self:updateBar(currLayout.targetTargetHealth, maxVal, curVal, curFrac, color)            
            end
            if currLayout.targetTargetPower ~= "e" then
                maxVal, curVal, curFrac = getPower("targettarget")
                color.r, color.g, color.b, color.hex = self:colorize("targetTarget", UnitPowerType("targetarget"),  curFrac)
                self:updateBar(currLayout.targetTargetPower, maxVal, curVal, curFrac, color)            
            end
            self:updateText("targetTarget", p.show.targetTargetName)
        else
            if currLayout.targetTargetHealth ~= "e" then
                self:updateBar(currLayout.targetTargetHealth, 0, 0, 0)            
            end
            if currLayout.targetTargetPower ~= "e" then
                self:updateBar(currLayout.targetTargetPower, 0, 0, 0)            
            end
            DHUD2_targetTarget_Text:Hide()
        end
    end
    self:updateAlpha()
end

function DHUD2:UNIT_HAPPINESS(who)
    if who == "pet" then
        local happ = GetPetHappiness()
        local health = UnitHealth(who)
        self:updatePetIcon(health, happ)
    end
end

function DHUD2:UNIT_PET(who)
    local maxVal, curVal, curFrac
    local color = {r=0,g=0,b=0,hex=""}
    local p = self.db.profile
    if who == "player" then
        hasPet = UnitExists("pet")
        DHUD2_lpi:Hide()
        DHUD2_lpo:Hide()
        DHUD2_rpi:Hide()
        DHUD2_rpo:Hide()
        if hasPet and p.show.pet then
            maxVal, curVal, curFrac, regen = getHealth("pet")
            color.r, color.g, color.b, color.hex = self:colorize("pet", "Health", curVal)
            self:updateBar(currLayout.petHealth, maxVal, curVal, curFrac)
            maxVal, curVal, curFrac, regen = getPower("pet")
            local typ = UnitPowerType(who)
            color.r, color.g, color.b, color.hex = self:colorize("pet", typ, curVal)
            self:updateBar(currLayout.petPower, maxVal, curVal, curFrac)
            local happ = GetPetHappiness()
            self:updatePetIcon(curVal, happ)
        else
            self:updatePetIcon(0, false)
            self:updateBar(currLayout.petHealth, 0, 0, 0)
            self:updateBar(currLayout.petPower, 0, 0, 0)
        end
    end
    self:updateAlpha()
end

function DHUD2:RAID_TARGET_UPDATE()
    self:updateTargetIcon()
end

function DHUD2:PARTY_LOOT_METHOD_CHANGED()
    local _, master = GetLootMethod()
    if master == 0 then
        DHUD2_playerLooter:Show()
    else
        DHUD2_playerLooter:Hide()
    end
end

function DHUD2:PARTY_MEMBERS_CHANGED()
    if GetNumPartyMembers() > 0 then
        if IsPartyLeader() then
            DHUD2_playerLeader:Show()
        else
            DHUD2_playerLeader:Hide()
        end
    else
        DHUD2_playerLeader:Hide()
        DHUD2_playerLooter:Hide()
    end
    
end


function DHUD2:UNIT_SPELLCAST_START(who)
    local p = self.db.profile
    if p.show.casting and who == "player" then        
        local _, _, _, _, startTime, endTime = UnitCastingInfo(who)
        --this.startTime  = startTime / 1000
        finishTime = endTime / 1000
        spellDuration = (endTime - startTime)/1000
        delayTime = 0
        isCasting = true
        isChanneling = false
        DHUD2_casting:Show()
        DHUD2_flash:Hide()
        updateCastText(spellDuration)
        DHUD2_castTime_Text:Show()
        self:updateAlpha()
        castTimer = 0
        DHUD2_casting:SetScript("OnUpdate", self.OnUpdateCast)
    end
end

function DHUD2:UNIT_SPELLCAST_CHANNEL_START(who)
    if self.db.profile.show.casting and who == "player" then
        local _, _, _, _, startTime, endTime = UnitChannelInfo(who)
        --this.startTime  = startTime / 1000;
        finishTime = endTime / 1000;
        spellDuration   = (endTime - startTime) / 1000
        isCasting = false
        isChanneling = true
        delayTime = 0
        setBarHeight("casting", 1)
        DHUD2_casting_Texture:SetVertexColor(self:colorize("casting", "channel", 0))        
        DHUD2_casting:Show()
        DHUD2_flash:Hide()
        updateCastText(spellDuration)
        DHUD2_castTime_Text:Show()
        self:updateAlpha()
        castTimer = 0
        DHUD2_casting:SetScript("OnUpdate", self.OnUpdateCast)
    end
end

function DHUD2:UNIT_SPELLCAST_STOP(who)
    local p = self.db.profile
    if p.show.casting and who == "player" then
        if isCasting then
            isCasting = false
            isChanneling = false
            flashCastBar()
            fadeoutCastBar()
            DHUD2_casting_Texture:SetVertexColor(0, 1, 0)
            setBarHeight("casting", 1)
            updateCastText(0)
            DHUD2_castTime_Text:Hide()
            DHUD2_castDelay_Text:Hide()
        end
    end
end

function DHUD2:UNIT_SPELLCAST_CHANNEL_STOP(who)
    if self.db.profile.show.casting and who == "player" then
        if isChanneling then
            isCasting = false
            isChanneling = false
            flashCastBar()
            fadeoutCastBar()
            setBarHeight("casting", 1)
            updateCastText(0)
            DHUD2_castTime_Text:Hide()
            DHUD2_castDelay_Text:Hide()
        end
    end
end

function DHUD2:OnSpellFail(who)        
    if self.db.profile.show.casting and who == "player" then
        if not barFadeOut then
            DHUD2_casting_Texture:SetVertexColor(1, 0, 0)
            setBarHeight("casting", 0)
            isCasting = false
            isChanneling = false
            fadeoutCastBar()
            updateCastText(0)
            DHUD2_castTime_Text:Hide()
            DHUD2_castDelay_Text:Hide()          
        end        
    end
end
    
function DHUD2:UNIT_SPELLCAST_DELAYED(who)
    if self.db.profile.show.casting and who == "player" then
        local _, _, _, _, startTime, endTime = UnitCastingInfo(who)
        local delay = endTime / 1000 - finishTime
        --this.startTime = this.startTime + delay;
        finishTime = endTime
        delayTime = delayTime + delay
        self:Debug("UNIT_SPELLCAST_DELAYED", delayTime)
        spellDuration = spellDuration + delay
        --local time = GetTime();
        --if (time > this.endTime) then
        --    time = this.endTime
        --end
    end
end
   
function DHUD2:UNIT_SPELLCAST_CHANNEL_UPDATE(who)
    if self.db.profile.show.casting and who == "player" then
        local _, _, _, _, startTime, endTime = UnitCastingInfo(who)
        
        --local origDuration = this.endTime - this.startTime
        --local elapsedTime = GetTime() - this.startTime;
--                delayTime = (origDuration - elapsedTime) - (who/1000);
--                this.endTime = GetTime() + (who / 1000);
        -- hack
        --if endTime == nil then
        --    endTime = 0;
        --end
        local delay = endTime / 1000 - finishTime
        delayTime = delayTime + delay
        self:Debug ("CHANNEL_UPDATE", delayTime)
        spellDuration = spellDuration + delay
   end
end

function DHUD2:UNIT_DISPLAYPOWER(who)
    self:Debug("UNIT_DISPLAYPOWER", who)
    local p = self.db.profile
    local maxVal, curVal, curFrac
    local color = {r=0,g=0,b=0,hex=""}
    if who == "player" then
        if hasPet then 
            return
        elseif isDruid and SmartyCat then
            if UnitPowerType("player") == 0 then
                self:updateBar(currLayout.druidMana, 0, 0, 0)
                self:UnregisterEvent("SmartyCat_ManaUpdated")
            else
                _G["DHUD2_"..currLayout.druidMana]:Show()
                self:RegisterEvent("SmartyCat_ManaUpdated", "OnFormUpdateMana")
                curVal, maxVal = SmartyCat.GetCurrMana(), SmartyCat.GetMaxMana()
                color.r, color.g, color.b, color.hex = self:colorize("player", 0, curVal/maxVal)
                self:updateBar(currLayout.druidMana, maxVal, curVal, curVal/maxVal, color)
            end
        end
        maxVal, curVal, curFrac, need = getPower("player")
        color.r, color.g, color.b, color.hex = self:colorize("player", UnitPowerType("player"), curFrac)
        self:updateBar(currLayout.playerPower, maxVal, curVal, curFrac, color)
        self:updateAlpha()
    end
end


function DHUD2:OnFormUpdateMana(curVal, maxVal)
    if hasPet then
        return
    end
    local color = {r=0,g=0,b=0,hex=""}
    color.r, color.g, color.b, color.hex = self:colorize("player", 0, curVal/maxVal)
    self:updateBar(currLayout.druidMana, maxVal, curVal, curVal/maxVal, color)
    self:updateAlpha()
end




function DHUD2_playerDropDownInit()
    UnitPopup_ShowMenu( DHUD2_playerDropDown, "SELF", "player" );
end

function DHUD2_targetDropDownInit()
    local menu = nil
    if (UnitIsEnemy("target", "player")) then
        return
    end
    if (UnitIsUnit("target", "player")) then
        menu = "SELF"
    elseif (UnitIsUnit("target", "pet")) then
        menu = "PET"
    elseif (UnitIsPlayer("target")) then
        if (UnitInParty("target")) then
            menu = "PARTY"
        else
            menu = "PLAYER"
        end
    end
    if menu then
        UnitPopup_ShowMenu( DHUD2_targetDropDown, menu, "target" );
    end
end
